/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.ssh;

import com.monica.javacom.interfaces.*;
import com.monica.javacom.ssh.*;
import com.monica.javacom.util.*;

import java.math.*;
import java.util.*;
import java.io.*;

/** This is the main class for the MCJ SSH2 layer. It provides functionality for
    the transport, authentication and connection layers of SSH2. This class is
    compliand to the MJC layer hierarchy and can be placed between the MJCTerminal and
    the MJCTransport layers. It provides both MJCProtocol and MJCCiphering functionality
    in the means of the MJC layer hierarchy.<br><br>
    
    <ul>
      <li>Implemented issues from the <b>TRANSPORT</b> layer:<br>
        <ul>
	  <li><b>kexinit</b>
	    <ul>
	      <li>diffie-hellman-group1-sha1
	      <li>dss
	    </ul>
          <li><b>ciphering</b>
            <ul>
	      <li>arcfour
	      <li>3des-cbc
      	      <li>none
	    </ul>
          <li><b>mac</b>
            <ul>
	      <li>hmac-sha1
              <li>hmac-md5 !beta!
	      <li>none
            </ul>
	  <li><b>compression</b>
            <ul>
	      <li>none
	    </ul>
	  <li><b>Implemented SSH2 Transport Protocol Layer messages</b>
	    <ul>
	      <li>SSH_TRANSPORT_MSG_KEXINIT
	      <li>SSH_TRANSPORT_MSG_KEXDH_REPLY
	      <li>SSH_TRANSPORT_MSG_NEWKEYS
	      <li>SSH_TRANSPORT_MSG_SERVICE_ACCEPT
	      <li>SSH_TRANSPORT_MSG_DISCONNECT
	    </ul>  	      
	</ul><br><br><br>
	
	
      <li>Implemented issues from the <b>AUTHERTICATION</b> layer:<br>	
        <ul>
	  <li><b>user authentication</b>
	    <ul>
	      <li>password	      
	    </ul>
	  <li><b>Implemented SSH2 Userauth Protocol Layer messages</b>    
	    <ul>
	      <li>SSH_USERAUTH_MSG_USERAUTH_SUCCESS
	      <li>SSH_USERAUTH_MSG_USERAUTH_FAILURE
	    </ul>  
	</ul><br><br><br>
	
		        
      <li>Implemented issues from the <b>CONNECTION</b> layer:<br>	      
        <ul>
	  <li><b>Implemented SSH2 Connection Protocol Layer messages</b>    
	    <ul>
              <li>SSH_CONNECTION_MSG_CHANNEL_OPEN_CONFIRMATION
              <li>SSH_CONNECTION_MSG_CHANNEL_SUCCESS
              <li>SSH_CONNECTION_MSG_CHANNEL_DATA
              <li>SSH_CONNECTION_MSG_CHANNEL_WINDOW_ADJUST
              <li>SSH_CONNECTION_MSG_CHANNEL_CLOSE
              <li>SSH_CONNECTION_MSG_CHANNEL_REQUEST
	    </ul>  
	</ul>    
    </ul>
      
*/
public class MJCSSHTransportLayer implements MJCReader, MJCWriter, MJCReadable, MJCWriteable, MJCConfigureable, MJCCloser, Runnable {

  // *******************************************************************************************
  // Constants
  // *******************************************************************************************

  private static final String CONF_KEX_KA="SSH2_SUPPORTED_KEXINIT_KEX_ALGORITHMS";
  private static final String CONF_KEX_SHKA="SSH2_SUPPORTED_KEXINIT_SERVER_HOST_KEY_ALGORITHMS";
  private static final String CONF_KEX_EA_CTS="SSH2_SUPPORTED_KEXINIT_ENCRYPTION_ALGORITHMS_CLIENT_TO_SERVER";
  private static final String CONF_KEX_EA_STC="SSH2_SUPPORTED_KEXINIT_ENCRYPTION_ALGORITHMS_SERVER_TO_CLIENT";
  private static final String CONF_KEX_MA_CTS="SSH2_SUPPORTED_KEXINIT_MAC_ALGORITHMS_CLIENT_TO_SERVER";
  private static final String CONF_KEX_MA_STC="SSH2_SUPPORTED_KEXINIT_MAC_ALGORITHMS_SERVER_TO_CLIENT";
  private static final String CONF_KEX_CO_CTS="SSH2_SUPPORTED_KEXINIT_COMPRESSION_ALGORITHMS_CLIENT_TO_SERVER";
  private static final String CONF_KEX_CO_STC="SSH2_SUPPORTED_KEXINIT_COMPRESSION_ALGORITHMS_SERVER_TO_CLIENT";
  private static final String CONF_KEX_LA_CTS="SSH2_SUPPORTED_KEXINIT_LANGUAGES_CLIENT_TO_SERVER";
  private static final String CONF_KEX_LA_STC="SSH2_SUPPORTED_KEXINIT_LANGUAGES_SERVER_TO_CLIENT";

  private static final String NAME="MJCSSHTransportLayer";
  private static final String VERSION="0.2";
  private static final String SSH_VERSION="SSH-2.00-MJC"+VERSION;

  private static final int STAT_INIT=0;
  private static final int STAT_SSH=1;

  private static final String CRLF=""+((char)13)+((char)10);

  private static final String UO_USERNAME="Username:";
  private static final String UO_PASSWORD="Password:";
  private static final String UO_INCORRECTLOGIN="Login incorrect."+CRLF;
  private static final String UO_CANNOTCONTINUE="Cannot continue login attempt."+CRLF;

  private static final int UI_USERNAME_LENGTH=40;
  private static final int UI_PASSWORD_LENGTH=40;

  // Key Exchange Constants

  private static final int KEV_x_keybits=1024;

  // Channel

  private static final int CHANNEL_WINDOW_SIZE=100000;
  private static final int CHANNEL_MAX_PACKET_LENGTH=10000;

  // login attempt

  private static final int LA_REQ=0;
  private static final int LA_REQ_ACCEPT=1;

  // *******************************************************************************************
  // Variables
  // *******************************************************************************************

  private boolean wasInitSSH;
  private int status;

  private boolean bTransportProtocolReady, bAuthenticationProtocolReady;

  private MJCSSHTransportPackageTokenizer tokenizer;
  private MJCSSHIO io;
  private MJCTransport transport;
  private MJCWindow window;

  private Random rnd;

  // Protocol negotiation

  // possible values
  String PP_KA, PP_SHKA, PP_CI_CTS, PP_CI_STC, PP_IN_CTS, PP_IN_STC, PP_CO_CTS, PP_CO_STC, PP_LA_CTS, PP_LA_STC;

  String PN_KA; // negotiated kex algorithm
  String PN_SHKA; // negotiated server host key algorithm
  String PN_CI_CTS; // negotiated cipher suite client to server
  String PN_CI_STC; // negotiated cipher suite server to client
  String PN_IN_CTS; // negotiated integrity suite client to server
  String PN_IN_STC; // negotiated integrity suite server to client
  String PN_CO_CTS; // negotiated compression suite client to server
  String PN_CO_STC; // negotiated compression suite server to client
  String PN_LA_CTS; // negotiated language suite client to server
  String PN_LA_STC; // negotiated compression suite server to client

  // Key Exchange Variables

  BigInteger KEV_p, KEV_q, KEV_g, KEV_e, KEV_f, KEV_x, KEV_K;
  String KEV_V_C, KEV_V_S, KEV_I_C, KEV_I_S, KEV_K_S, KEV_H;
  BigInteger DSS_p, DSS_q, DSS_g, DSS_y, DSS_r, DSS_s;

  String OKE_IV_CTS, OKE_IV_STC, OKE_EK_CTS, OKE_EK_STC, OKE_IK_CTS, OKE_IK_STC; // output from keyexchange

  String sessionID;

  boolean kexDoneCTS, kexDoneSTC;

  // Algorithms

  private MJCSSHHashAlgorithm hash;
  private MJCSSHSHA1 sha1;

  // Channels

  private int clientChannelNo, serverChannelNo;
  private int channelWindowCTS, channelWindowSTC;
  private int maxPacketSizeCTS, maxPacketSizeSTC;

  private boolean connected;

  // MJC Layer Hierarchy

  private MJCStringBuffer readableBuf, writeableBuf;

  private MJCReadable readable;
  private MJCWriteable writeable;
  private MJCConfiguration configuration;
  private MJCReader reader;

  // User authentication

  private boolean UII_username, UII_password; // user input indication
  private String UIS_username, UIS_password; // user input strings

  // Thread management

  private boolean threadSuspended;
  private boolean threadRunning;
  private Thread myThread;

  // *******************************************************************************************
  // Initialization
  // *******************************************************************************************

  // Constructor

  public MJCSSHTransportLayer() {
    status=STAT_INIT;
    wasInitSSH=false;
    tokenizer=new MJCSSHTransportPackageTokenizer();

    tokenizer.setCipherAlgorithm(new MJCSSHNoneCiphering(),MJCSSHConstants.SSH_TOKENIZER_CTS);
    tokenizer.setCipherAlgorithm(new MJCSSHNoneCiphering(),MJCSSHConstants.SSH_TOKENIZER_STC);
    tokenizer.setMACAlgorithm(new MJCSSHNoneMAC(),MJCSSHConstants.SSH_TOKENIZER_CTS);
    tokenizer.setMACAlgorithm(new MJCSSHNoneMAC(),MJCSSHConstants.SSH_TOKENIZER_STC);
    tokenizer.setCompressionAlgorithm(new MJCSSHNoneCompression(),MJCSSHConstants.SSH_TOKENIZER_CTS);
    tokenizer.setCompressionAlgorithm(new MJCSSHNoneCompression(),MJCSSHConstants.SSH_TOKENIZER_STC);
    io=new MJCSSHIO();
    rnd=new Random();

    kexDoneSTC=false;
    kexDoneCTS=false;

    clientChannelNo=1;
    connected=false;

    readableBuf=new MJCStringBuffer(100000);
    writeableBuf=new MJCStringBuffer(100000);

    UII_password=false;
    UII_username=false;
    UIS_username=new String();
    UIS_password=new String();

    bTransportProtocolReady=false;
    bAuthenticationProtocolReady=false;

    // init p, g, q for diffie-hellman-group1-sha1

    String temp=new String();
    temp=temp.concat(new String("179769313486231590770839156793787453197860296048756011706444"));
    temp=temp.concat(new String("423684197180216158519368947833795864925541502180565485980503"));
    temp=temp.concat(new String("646440548199239100050792877003355816639229553136239076508735"));
    temp=temp.concat(new String("759914822574862575007425302077447712589550957937778424442426"));
    temp=temp.concat(new String("617334727629299387668709205606050270810842907692932019128194"));
    temp=temp.concat(new String("467627007"));

    KEV_p=new BigInteger(temp);
    KEV_g=new BigInteger("2");
    KEV_q=(KEV_p.subtract(new BigInteger("1"))).divide(KEV_g);

    sha1=new MJCSSHSHA1();
    
    OKE_IV_CTS=new String();
    OKE_IV_STC=new String();
    OKE_EK_CTS=new String();
    OKE_EK_STC=new String();
    OKE_IK_CTS=new String();
    OKE_IK_STC=new String();

    threadRunning=true;
    threadSuspended=true;
    myThread=new Thread(this);
    myThread.start();

  }

  // *******************************************************************************************
  // Runnable
  // *******************************************************************************************

  /** Entry point for the reader thread.
  */
  public void run() {
    while (threadRunning) {
      synchronized(this) {
        while (threadSuspended) {
          try {
            wait();
          }
          catch (Exception e) {
            //--
          }
        }
        threadSuspended=true;
      }
      process(readable.getInput());
    }
  }

  // *******************************************************************************************
  // MJCCloser
  // *******************************************************************************************

  public String getCloserName() {
    return NAME;
  }

  public String getCloserVersion() {
    return VERSION;
  }

  public void closeApplication() {
    closeConnection();
    transport.closeConnections();
  }

  public void setTransport(MJCTransport transport) {
    this.transport=transport;
  }

  public void setWindow(MJCWindow window) {
    this.window=window;
  }

  // *******************************************************************************************
  // MJCConfigureable
  // *******************************************************************************************

  public void setConfiguration(MJCConfiguration configuration) {
    this.configuration=configuration;
    rereadConfig();
  }

  public MJCConfiguration getConfiguration() {
    return configuration;
  }

  private void rereadConfig() {

    // kexinit algorithm negotiatioin settings

    PP_KA=configuration.getString(CONF_KEX_KA, MJCSSHConstants.SUPPORTED_KEXINIT_KEX_ALGORITHMS);
    PP_SHKA=configuration.getString(CONF_KEX_SHKA, MJCSSHConstants.SUPPORTED_KEXINIT_SERVER_HOST_KEY_ALGORITHMS);
    PP_CI_CTS=configuration.getString(CONF_KEX_EA_CTS, MJCSSHConstants.SUPPORTED_KEXINIT_ENCRYPTION_ALGORITHMS_CLIENT_TO_SERVER);
    PP_CI_STC=configuration.getString(CONF_KEX_EA_STC, MJCSSHConstants.SUPPORTED_KEXINIT_ENCRYPTION_ALGORITHMS_SERVER_TO_CLIENT);
    PP_IN_CTS=configuration.getString(CONF_KEX_MA_CTS, MJCSSHConstants.SUPPORTED_KEXINIT_MAC_ALGORITHMS_CLIENT_TO_SERVER);
    PP_IN_STC=configuration.getString(CONF_KEX_MA_STC, MJCSSHConstants.SUPPORTED_KEXINIT_MAC_ALGORITHMS_SERVER_TO_CLIENT);
    PP_CO_CTS=configuration.getString(CONF_KEX_CO_CTS, MJCSSHConstants.SUPPORTED_KEXINIT_COMPRESSION_ALGORITHMS_CLIENT_TO_SERVER);
    PP_CO_STC=configuration.getString(CONF_KEX_CO_STC, MJCSSHConstants.SUPPORTED_KEXINIT_COMPRESSION_ALGORITHMS_SERVER_TO_CLIENT);
    PP_LA_CTS=configuration.getString(CONF_KEX_LA_CTS, MJCSSHConstants.SUPPORTED_KEXINIT_LANGUAGES_CLIENT_TO_SERVER);
    PP_LA_STC=configuration.getString(CONF_KEX_LA_STC, MJCSSHConstants.SUPPORTED_KEXINIT_LANGUAGES_SERVER_TO_CLIENT);
  }

  // *******************************************************************************************
  // MJCReadable
  // *******************************************************************************************

  public String getReadableName() {
    return NAME;
  }

  public String getReadableVersion() {
    return VERSION;
  }

  public synchronized int getAvailable() {
    return readableBuf.available();
  }

  public synchronized String getInput() {
    return readableBuf.emptyBuffer();
  }

  public boolean isReadableConnected() {
    return connected;
  }

  public void setReader(MJCReader reader) {
    this.reader=reader;
  }

  public MJCReader getReader() {
    return reader;
  }

  // *******************************************************************************************
  // MJCWriteable
  // *******************************************************************************************

  public String getWriteableName() {
    return NAME;
  }

  public String getWriteableVersion() {
    return VERSION;
  }

  public boolean write(char c) {
    if ((!UII_username) && (!UII_password)) {
      if (connected) {
        return writeableBuf.addToBuffer(c);
      }
      else {
        return false;
      }
    }
    else {
      char ca[]=new char[1];
      ca[0]=c;
      return parseUI_UP(new String(ca));
    }
  }

  public boolean write(char[] ca) {
    if ((!UII_username) && (!UII_password)) {
      if (connected) {
        return writeableBuf.addToBuffer(ca);
      }
      else {
        return false;
      }
    }
    else {
      return parseUI_UP(new String(ca));
    }
  }

  public boolean write(String s) {
    if ((!UII_username) && (!UII_password)) {
      if (connected) {
        return writeableBuf.addToBuffer(s);
      }
      else {
        return false;
      }
    }
    else {
      return parseUI_UP(s);
    }
  }

  public void flush() {
    if ((!UII_username) && (!UII_password)) {
      if (connected) {
        sendWriteableBuffer();
      }
    }
  }

  public String getLastWriteableError() {
    return new String();
  }

  public boolean isWriteableConnected() {
    return connected;
  }

  // *******************************************************************************************
  // MJCReader
  // *******************************************************************************************

  public String getReaderName() {
    return NAME;
  }

  public String getReaderVersion() {
    return VERSION;
  }

  public synchronized void readEvent(MJCReadable readable) {
    if (readable.getAvailable()>0) {
      threadSuspended=false;
      notify();
    }
  }

  public synchronized void connectionClosedEvent(MJCReadable readable) {
    debugOutput("Connection closed by other side.");
    connected=false;
    writeToReader("Connection closed."+CRLF);
    threadRunning=false;
    threadSuspended=false;
    notify();
  }

  public synchronized void errorEvent(String errorDesc, MJCReadable readable) {
    connected=false;
    writeToReader("Error occured ("+errorDesc+"). Connection closed."+CRLF);
    closeConnection();
    threadRunning=false;
    threadSuspended=false;
    notify();
  }

  public synchronized void setReadable(MJCReadable readable) {
    this.readable=readable;
  }

  public MJCReadable getReadable() {
    return readable;
  }

  // *******************************************************************************************
  // MJCWriter
  // *******************************************************************************************

  public String getWriterName() {
    return NAME;
  }

  public String getWriterVersion() {
    return VERSION;
  }

  public void setWriteable(MJCWriteable writeable) {
    this.writeable=writeable;
  }

  public MJCWriteable getWriteable() {
    return writeable;
  }

  // *******************************************************************************************
  // Main Message Processing
  // *******************************************************************************************

  /** Processes the stream represented as a String received from the upper MJC layer.
      @param String msg: the stream String
  */
  private void process(String msg) {

//    System.out.println("in="+msg);
//    System.out.println();

    String cur_msg;
    MJCSSHTransportPackage pack;
    BigInteger msg_type;
    int msg_type_int;
    String temp;

    switch (status) {

      case STAT_INIT:

        while (msg.length()>0) {

          //writeToReader("init:"+msg+CRLF);

          cur_msg=io.getNextToken(msg, MJCSSHConstants.SEP_CRLF);

          if ((cur_msg.length()>2) && (cur_msg.substring(0,3).equals("SSH"))) {
            KEV_V_S=cur_msg;
            wasInitSSH=true;
            writeable.write(SSH_VERSION+((char)13)+((char)10));
            KEV_V_C=SSH_VERSION;
            writeable.flush();
            msg=io.delNextToken(msg, MJCSSHConstants.SEP_CRLF);
          }
          else {
            if (wasInitSSH) {
              status=STAT_SSH;
              process(msg);
              break;
            }
            else {
              debugOutput(cur_msg);
            }
          }
        }
        break;

      case STAT_SSH:

        //writeToReader("ssh"+CRLF);

        try {
          tokenizer.arrived(msg);
        }
        catch (Exception e) {
          writeToReader(CRLF);
          writeToReader("Exception occured while SSH2 processing: "+e.getMessage());
          closeConnection();
          return;
        }

        while (tokenizer.isAvailable()) {
          pack=tokenizer.getNextPackage();
          msg_type=io.rByte(pack.payload);
          msg_type_int=msg_type.intValue();

          //writeToReader("in="+msg_type_int);

          switch (msg_type_int) {

            // Transport
	    case MJCSSHConstants.SSH_TRANSPORT_MSG_IGNORE :
	      debugOutput("ignore: "+pack.payload.substring(1));
	      break;
            case MJCSSHConstants.SSH_TRANSPORT_MSG_KEXINIT :
              processKexInitMsg(pack);
              break;
            case MJCSSHConstants.SSH_TRANSPORT_MSG_KEXDH_REPLY :
              processKexDHReplyMsg(pack);
              break;
            case MJCSSHConstants.SSH_TRANSPORT_MSG_NEWKEYS :
              processNewKeysMsg(pack);
              break;
            case MJCSSHConstants.SSH_TRANSPORT_MSG_SERVICE_ACCEPT :
              processServiceAcceptMsg(pack);
              break;
            case MJCSSHConstants.SSH_TRANSPORT_MSG_DISCONNECT :
              processDisconnectMsg(pack);
              break;

            // Authentication
            case MJCSSHConstants.SSH_USERAUTH_MSG_USERAUTH_SUCCESS :
              processUserAuthSuccessMsg(pack);
              break;
            case MJCSSHConstants.SSH_USERAUTH_MSG_USERAUTH_FAILURE :
              processUserAuthFailureMsg(pack);
              break;

            // Connection
            case MJCSSHConstants.SSH_CONNECTION_MSG_CHANNEL_OPEN_CONFIRMATION :
              processChannelOpenConfirmationMsg(pack);
              break;
            case MJCSSHConstants.SSH_CONNECTION_MSG_CHANNEL_SUCCESS :
              processChannelSuccessMsg(pack);
              break;
            case MJCSSHConstants.SSH_CONNECTION_MSG_CHANNEL_DATA :
              processChannelDataMsg(pack);
              break;
            case MJCSSHConstants.SSH_CONNECTION_MSG_CHANNEL_WINDOW_ADJUST :
              processWindowAdjustMsg(pack);
              break;
            case MJCSSHConstants.SSH_CONNECTION_MSG_CHANNEL_CLOSE :
              processChannelClosedMsg(pack);
              break;
            case MJCSSHConstants.SSH_CONNECTION_MSG_CHANNEL_REQUEST :
              processChannelRequestMsg(pack);
              break;

            // Other
            default:
              sendUnimplementedMsg(pack);
              break;

          }
        }
        break;
    }
  }

  // *******************************************************************************************
  // Transport protocol messages
  // *******************************************************************************************

  /** Replies an UNIMPLEMENTED message for a packet that contains a message type that
      is not implemented in this MJC client or is invalid.
      @param MJCSSHTransportPackage pack: the packet that is invalid or unimplemented
  */
  private void sendUnimplementedMsg(MJCSSHTransportPackage pack) {
    int msg_type_int=io.rByteInt(pack.payload);
    debugOutput("Unknown message type="+msg_type_int+CRLF);

    String temp=io.wByte(MJCSSHConstants.SSH_TRANSPORT_MSG_UNIMPLEMENTED);
    temp=temp.concat(io.wUint32(pack.sequenceNumber));
    send(temp);
  }

  /** Processes the KEX_INIT message received from the SSH2 server.
      @param MJCSSHTransportPackage pack: the KEX_INIT message
  */
  private void processKexInitMsg(MJCSSHTransportPackage pack) {

    Vector data;
    String kA, sHKA, eACTS, eASTC, mACTS, mASTC, cACTS, cASTC, lCTS, lSTC;
    String s_kA, s_sHKA, s_eACTS, s_eASTC, s_mACTS, s_mASTC, s_cACTS, s_cASTC, s_lCTS, s_lSTC;
    Boolean fKPF;
    String reply_payload;

    KEV_I_S=pack.payload;

    data=io.parsePayload(pack.payload,MJCSSHConstants.KEX_INIT_MSG_DATA_DESC);

    kA=(String)(data.elementAt(MJCSSHConstants.DI_KEXINIT_KEX_ALGORITHMS));
    sHKA=(String)(data.elementAt(MJCSSHConstants.DI_KEXINIT_SERVER_HOST_KEY_ALGORITHMS));
    eACTS=(String)(data.elementAt(MJCSSHConstants.DI_KEXINIT_ENCRYPTION_ALGORITHMS_CLIENT_TO_SERVER));
    eASTC=(String)(data.elementAt(MJCSSHConstants.DI_KEXINIT_ENCRYPTION_ALGORITHMS_SERVER_TO_CLIENT));
    mACTS=(String)(data.elementAt(MJCSSHConstants.DI_KEXINIT_MAC_ALGORITHMS_CLIENT_TO_SERVER));
    mASTC=(String)(data.elementAt(MJCSSHConstants.DI_KEXINIT_MAC_ALGORITHMS_SERVER_TO_CLIENT));
    cACTS=(String)(data.elementAt(MJCSSHConstants.DI_KEXINIT_COMPRESSION_ALGORITHMS_CLIENT_TO_SERVER));
    cASTC=(String)(data.elementAt(MJCSSHConstants.DI_KEXINIT_COMPRESSION_ALGORITHMS_SERVER_TO_CLIENT));
    lCTS=(String)(data.elementAt(MJCSSHConstants.DI_KEXINIT_LANGUAGES_CLIENT_TO_SERVER));
    lSTC=(String)(data.elementAt(MJCSSHConstants.DI_KEXINIT_LANGUAGES_SERVER_TO_CLIENT));
    fKPF=(Boolean)(data.elementAt(MJCSSHConstants.DI_KEXINIT_FIRST_KEY_PACKET_FOLLOWS));

    s_kA=io.findMatch(kA, PP_KA, 2);
    s_sHKA=io.findMatch(sHKA, PP_SHKA, 2);
    s_eACTS=io.findMatch(eACTS, PP_CI_CTS, 2);
    s_eASTC=io.findMatch(eASTC, PP_CI_STC, 2);
    s_mACTS=io.findMatch(mACTS, PP_IN_CTS, 2);
    s_mASTC=io.findMatch(mASTC, PP_IN_STC, 2);
    s_cACTS=io.findMatch(cACTS, PP_CO_CTS, 2);
    s_cASTC=io.findMatch(cASTC, PP_CO_STC, 2);
    s_lCTS=io.findMatch(lCTS, PP_LA_CTS, 2);
    s_lSTC=io.findMatch(lSTC, PP_LA_STC, 2);

//    debugOutput("s_mACTS="+s_mACTS);

    try {
      PN_KA=io.chosenMatch(s_kA);
      PN_SHKA=io.chosenMatch(s_sHKA);
      PN_CI_CTS=io.chosenMatch(s_eACTS);
      PN_CI_STC=io.chosenMatch(s_eASTC);
      PN_CO_CTS=io.chosenMatch(s_cACTS);
      PN_CO_STC=io.chosenMatch(s_cASTC);
      PN_IN_CTS=io.chosenMatch(s_mACTS);
      PN_IN_STC=io.chosenMatch(s_mASTC);
    }
    catch (Exception e) {
      debugOutput("Could not negotiate protocol algorithms.");
      writeToReader("Could not negotiate protocol algorithms.");
      closeConnection();
      return;
    }

//    debugOutput("PN_IN_CTS="+PN_IN_CTS);

    reply_payload=new String();
    reply_payload=reply_payload.concat(io.wByte(MJCSSHConstants.SSH_TRANSPORT_MSG_KEXINIT));
    reply_payload=reply_payload.concat((String)(data.elementAt(MJCSSHConstants.DI_KEXINIT_COOKIE)));
    reply_payload=reply_payload.concat(io.wString(s_kA));
    reply_payload=reply_payload.concat(io.wString(s_sHKA));
    reply_payload=reply_payload.concat(io.wString(s_eACTS));
    reply_payload=reply_payload.concat(io.wString(s_eASTC));
    reply_payload=reply_payload.concat(io.wString(s_mACTS));
    reply_payload=reply_payload.concat(io.wString(s_mASTC));
    reply_payload=reply_payload.concat(io.wString(s_cACTS));
    reply_payload=reply_payload.concat(io.wString(s_cASTC));
    reply_payload=reply_payload.concat(io.wString(s_lCTS));
    reply_payload=reply_payload.concat(io.wString(s_lSTC));
    reply_payload=reply_payload.concat(io.wBoolean(true));
    reply_payload=reply_payload.concat(io.wUint32(0));

    KEV_I_C=reply_payload;

    send(reply_payload);

    sendKexDHInitMsg();
  }

  /** Client side KEXDH_INIT message creation.
  */
  private void sendKexDHInitMsg() {

    KEV_x=new BigInteger(KEV_x_keybits,rnd);
    KEV_e=KEV_g.modPow(KEV_x, KEV_p);

    String kinit=io.wByte(MJCSSHConstants.SSH_TRANSPORT_MSG_KEXDH_INIT);
    kinit=kinit.concat(io.wMpint(KEV_e));

    send(kinit);
  }

  /** Processes the KEXDH_REPLY message from the SSH2 server.
      @param MJCSSHTransportPackage pack: the KEX_INIT message
  */
  private void processKexDHReplyMsg(MJCSSHTransportPackage pack) {
    Vector data, dss_blob;
    String signature;

    // KEXDH_REPLY

    data=io.parsePayload(pack.payload,MJCSSHConstants.KEXDH_REPLY_MSG_DATA_DESC);

    KEV_K_S=(String)(data.elementAt(MJCSSHConstants.DI_KEXDH_REPLY_K_S));
    KEV_f=(BigInteger)(data.elementAt(MJCSSHConstants.DI_KEXDH_REPLY_F));
    signature=(String)(data.elementAt(MJCSSHConstants.DI_KEXDH_REPLY_SIGNATURE));

    String keyFormat=io.rString(KEV_K_S);

    if (!keyFormat.equals(PN_SHKA)) {
      debugOutput("Key format not the same.");
      closeConnection();
      return;
    }

    if (keyFormat.equals("ssh-dss")) {
      processSSHDSSKexDHReply(signature);
    }
    else {
      sendUnimplementedMsg(pack);
      closeConnection();
    }
  }

  /** Processes SSH-DSS key format.
      @param String signature: the signature for the DSS check
  */
  private void processSSHDSSKexDHReply(String signature) {

    Vector data_dss, dss_blob;
    String blob;

//    debugOutput("1");

    data_dss=io.parsePayload(KEV_K_S,MJCSSHConstants.KEXDH_REPLY_SSH_DSS_MSG_DATA_DESC);
    DSS_p=(BigInteger)(data_dss.elementAt(MJCSSHConstants.DI_KEXDH_REPLY_SSH_DSS_P));
    DSS_q=(BigInteger)(data_dss.elementAt(MJCSSHConstants.DI_KEXDH_REPLY_SSH_DSS_Q));
    DSS_g=(BigInteger)(data_dss.elementAt(MJCSSHConstants.DI_KEXDH_REPLY_SSH_DSS_G));
    DSS_y=(BigInteger)(data_dss.elementAt(MJCSSHConstants.DI_KEXDH_REPLY_SSH_DSS_Y));

//    debugOutput("2");

//    dumpString(signature, System.out);

    if (signature.length()>40) { // standard
      dss_blob=io.parsePayload(signature,MJCSSHConstants.KEXDH_REPLY_DSS_BLOB_MSG_DATA_DESC);
      blob=(String)(dss_blob.elementAt(MJCSSHConstants.DI_KEXDH_REPLY_DSS_BLOB_DATA));
    }
    else { // fuck what the heck is that
      blob=signature;
    }

//    debugOutput("3");

    DSS_r=io.stringAsBigInteger(blob.substring(0,20));
    DSS_s=io.stringAsBigInteger(blob.substring(20));

    // Computation

    KEV_K=KEV_f.modPow(KEV_x, KEV_p);

    // DSS Verification

    String concatenated=new String(io.wString(KEV_V_C));
    concatenated=concatenated.concat(io.wString(KEV_V_S));
    concatenated=concatenated.concat(io.wString(KEV_I_C));
    concatenated=concatenated.concat(io.wString(KEV_I_S));
    concatenated=concatenated.concat(io.wString(KEV_K_S));
    concatenated=concatenated.concat(io.wMpint(KEV_e));
    concatenated=concatenated.concat(io.wMpint(KEV_f));
    concatenated=concatenated.concat(io.wMpint(KEV_K));

    KEV_H=sha1.hash(concatenated);

    if (DSSCheck(KEV_H, DSS_p, DSS_q, DSS_g, DSS_y, DSS_r, DSS_s)) {
      writeToReader("Host verified."+CRLF);
      sessionID=KEV_H;

      String temp=(io.wMpint(KEV_K)).concat(KEV_H);

      // IV CTS
      
      if (PN_CI_CTS.equals("arcfour")) {
      }
/*      else if (PN_CI_CTS.equals("aes128-cbc")) {
        OKE_IV_CTS=createKeyData(temp, "A", sessionID, 16, sha1);
      }*/
      else if (PN_CI_CTS.equals("3des-cbc")) {
        OKE_IV_CTS=createKeyData(temp, "A", sessionID, 8, sha1);
      }
      else if (PN_CI_CTS.equals("none")) {
      }
      
      // IV STC

      if (PN_CI_STC.equals("arcfour")) {
      }
/*      else if (PN_CI_STC.equals("aes128-cbc")) {
        OKE_IV_STC=createKeyData(temp, "B", sessionID, 16, sha1);
      }*/
      else if (PN_CI_STC.equals("3des-cbc")) {
        OKE_IV_STC=createKeyData(temp, "B", sessionID, 8, sha1);
      }
      else if (PN_CI_STC.equals("none")) {
      }
      
      // encryption CTS

      if (PN_CI_CTS.equals("arcfour")) {
        OKE_EK_CTS=createKeyData(temp, "C", sessionID, 16, sha1);
      }
/*      else if (PN_CI_CTS.equals("aes128-cbc")) {
        OKE_EK_CTS=createKeyData(temp, "C", sessionID, 16, sha1);
      }*/
      else if (PN_CI_CTS.equals("3des-cbc")) {
        OKE_EK_CTS=createKeyData(temp, "C", sessionID, 24, sha1);
      }
      else if (PN_CI_CTS.equals("none")) {
      }
      
      // encryption STC

      if (PN_CI_STC.equals("arcfour")) {
        OKE_EK_STC=createKeyData(temp, "D", sessionID, 16, sha1);
      }
/*      else if (PN_CI_STC.equals("aes128-cbc")) {
        OKE_EK_STC=createKeyData(temp, "D", sessionID, 16, sha1);
      }*/
      else if (PN_CI_STC.equals("3des-cbc")) {
        OKE_EK_STC=createKeyData(temp, "D", sessionID, 24, sha1);
      }
      else if (PN_CI_STC.equals("none")) {
      }
      
      // integrity CTS

      if (PN_IN_CTS.equals("hmac-sha1")) {
        OKE_IK_CTS=createKeyData(temp, "E", sessionID, 20, sha1);
      }
      else if (PN_IN_CTS.equals("hmac-md5")) {
        OKE_IK_CTS=createKeyData(temp, "E", sessionID, 16, sha1);
      }
      else if (PN_IN_CTS.equals("none")) {
      }
      
      // integrity STC

      if (PN_IN_STC.equals("hmac-sha1")) {
        OKE_IK_STC=createKeyData(temp, "F", sessionID, 20, sha1);
      }
      else if (PN_IN_STC.equals("hmac-md5")) {
        OKE_IK_STC=createKeyData(temp, "F", sessionID, 16, sha1);
      }
      else if (PN_IN_STC.equals("none")) {
      }
      
      debugOutput("A="+io.print(OKE_IV_CTS));
      debugOutput("B="+io.print(OKE_IV_STC));
      debugOutput("C="+io.print(OKE_EK_CTS));
      debugOutput("D="+io.print(OKE_EK_STC));
      debugOutput("E="+io.print(OKE_IK_CTS));
      debugOutput("F="+io.print(OKE_IK_STC));

      sendNewKeysMsg();
    }
    else {
      writeToReader("Host verification failed."+CRLF);
      closeConnection();
    }
  }

  /** Create arbitrary length key data as described in the SSH2 key data creation
      algorithm.
      @param String firstPart: the first, constant part of the iteration
      @param String variant: variant of the first round, e.g. "A"
      @param String session: the session ID
      @param int numBytes: how many bytes should the key data contain
      @param MJCSSHHashAlgorithm: the HASH algorithm to be used for creation of the key data
      @return String: the final key data result
  */
  private String createKeyData(String firstPart, String variant, String session, int numBytes,
                               MJCSSHHashAlgorithm h) {
    String res;
    res=h.hash(firstPart.concat(variant.concat(session)));
    while (res.length()<numBytes) {
      res=res.concat(h.hash(firstPart.concat(res)));
    }
    return res.substring(0,numBytes);
  }

  /** Performs the DSS check with the received keys.
      @param String DS_M: parameter M for the DSS check algorithm
      @param BigInteger DC_p: parameter p for the DSS check algorithm
      @param BigInteger DC_q: parameter q for the DSS check algorithm
      @param BigInteger DC_g: parameter g for the DSS check algorithm
      @param BigInteger DC_y: parameter y for the DSS check algorithm
      @param BigInteger DC_r: parameter r for the DSS check algorithm
      @param BigInteger DC_s: parameter s for the DSS check algorithm
      @return boolean: wether the DSS check was successful
  */
  private boolean DSSCheck(String DC_M, BigInteger DC_p, BigInteger DC_q, BigInteger DC_g,
                           BigInteger DC_y, BigInteger DC_r, BigInteger DC_s) {

    BigInteger DC_w=DC_s.modInverse(DC_q);
    BigInteger DC_hash=sha1.hashAsBigInteger(DC_M);
    BigInteger DC_u1=DC_w.multiply(DC_hash);
    DC_u1=DC_u1.mod(DC_q);
    BigInteger DC_u2=DC_r.multiply(DC_w);
    DC_u2=DC_u2.mod(DC_q);
    BigInteger t1, t2;
    t1=DC_g.modPow(DC_u1, DC_p);
    t2=DC_y.modPow(DC_u2, DC_p);
    BigInteger DC_v=((t1.multiply(t2)).mod(DC_p)).mod(DC_q);

    return (DC_v.equals(DC_r));
  }

  /** Sends the NEWKEYS message to the server.
  */
  private void sendNewKeysMsg() {
    String newKeys=io.wByte(MJCSSHConstants.SSH_TRANSPORT_MSG_NEWKEYS);
    kexDoneCTS=true;

    synchronized (this) {
      send(newKeys);

      tokenizer.setKeys(OKE_IV_CTS, OKE_EK_CTS, OKE_IK_CTS, MJCSSHConstants.SSH_TOKENIZER_CTS);

      // MAC
      
      if (PN_IN_CTS.equals("hmac-sha1")) {
        debugOutput("Using hmac-sha1 as MAC for CTS");
        tokenizer.setMACAlgorithm(new MJCSSHHMAC(new MJCSSHSHA1()), MJCSSHConstants.SSH_TOKENIZER_CTS);
      }
      else if (PN_IN_CTS.equals("hmac-md5")) {
        debugOutput("Using hmac-md5 as MAC for CTS");
        tokenizer.setMACAlgorithm(new MJCSSHHMAC(new MJCSSHMD5()), MJCSSHConstants.SSH_TOKENIZER_CTS);
      }
      else if (PN_IN_CTS.equals("none")) {
        debugOutput("Using none as MAC for CTS");
        tokenizer.setMACAlgorithm(new MJCSSHNoneMAC(), MJCSSHConstants.SSH_TOKENIZER_CTS);
      }
      
      // Ciphering

      if (PN_CI_CTS.equals("arcfour")) {
        debugOutput("Using arcfour as ciphering for CTS");
        tokenizer.setCipherAlgorithm(new MJCSSHArcfourCipher(),MJCSSHConstants.SSH_TOKENIZER_CTS);
      }
/*      else if (PN_CI_CTS.equals("aes128-cbc")) {
        debugOutput("Using aes128-cbc as ciphering for CTS");
        tokenizer.setCipherAlgorithm(new MJCSSHAES128CBC(),MJCSSHConstants.SSH_TOKENIZER_CTS);
      }*/
      else if (PN_CI_CTS.equals("3des-cbc")) {
        debugOutput("Using 3des-cbc as ciphering for CTS");
        tokenizer.setCipherAlgorithm(new MJCSSH3DESCBC(),MJCSSHConstants.SSH_TOKENIZER_CTS);
      }
      else if (PN_CI_CTS.equals("none")) {
        debugOutput("Using none as ciphering for CTS");
        tokenizer.setCipherAlgorithm(new MJCSSHNoneCiphering(),MJCSSHConstants.SSH_TOKENIZER_CTS);
      }
      
      // Compression
      
      if (PN_CO_CTS.equals("zlib")) {
        debugOutput("Using zlib as compression for CTS");
        tokenizer.setCompressionAlgorithm(new MJCSSHZLibCompression(), MJCSSHConstants.SSH_TOKENIZER_CTS);
      }  
      else if (PN_CO_CTS.equals("none")) {
        debugOutput("Using none as compression for CTS");
        tokenizer.setCompressionAlgorithm(new MJCSSHNoneCompression(), MJCSSHConstants.SSH_TOKENIZER_CTS);
      }  
        
      debugOutput("CTS Keys active");
    }
  }

  /** Processes the NEWKEYS message from the server.
      @param MJCSSHTransportPackage pack: the NEWKEYS message from the server
  */
  private void processNewKeysMsg(MJCSSHTransportPackage pack) {
    kexDoneSTC=true;

    tokenizer.setKeys(OKE_IV_STC, OKE_EK_STC, OKE_IK_STC, MJCSSHConstants.SSH_TOKENIZER_STC);
    
    // MAC

    if (PN_IN_STC.equals("hmac-sha1")) {
      debugOutput("Using hmac-sha1 as MAC for STC");
      tokenizer.setMACAlgorithm(new MJCSSHHMAC(new MJCSSHSHA1()), MJCSSHConstants.SSH_TOKENIZER_STC);
    }
    else if (PN_IN_STC.equals("hmac-md5")) {
      debugOutput("Using hmac-md5 as MAC for STC");
      tokenizer.setMACAlgorithm(new MJCSSHHMAC(new MJCSSHMD5()), MJCSSHConstants.SSH_TOKENIZER_STC);
    }
    else if (PN_IN_STC.equals("none")) {
      debugOutput("Using none as MAC for STC");
      tokenizer.setMACAlgorithm(new MJCSSHNoneMAC(), MJCSSHConstants.SSH_TOKENIZER_STC);
    }
    
    // Ciphering

    if (PN_CI_STC.equals("arcfour")) {
      debugOutput("Using arcfour as ciphering for STC");
      tokenizer.setCipherAlgorithm(new MJCSSHArcfourCipher(),MJCSSHConstants.SSH_TOKENIZER_STC);
    }
/*    else if (PN_CI_STC.equals("aes128-cbc")) {
      debugOutput("Using aes128-cbc as ciphering for STC");
      tokenizer.setCipherAlgorithm(new MJCSSHAES128CBC(),MJCSSHConstants.SSH_TOKENIZER_STC);
    }*/
    else if (PN_CI_STC.equals("3des-cbc")) {
      debugOutput("Using 3des-cbc as ciphering for STC");
      tokenizer.setCipherAlgorithm(new MJCSSH3DESCBC(),MJCSSHConstants.SSH_TOKENIZER_STC);
    }
    else if (PN_CI_STC.equals("none")) {
      debugOutput("Using none as ciphering for STC");
      tokenizer.setCipherAlgorithm(new MJCSSHNoneCiphering(),MJCSSHConstants.SSH_TOKENIZER_STC);
    }
    
    // Compression
      
    if (PN_CO_STC.equals("zlib")) {
      debugOutput("Using zlib as compression for STC");
      tokenizer.setCompressionAlgorithm(new MJCSSHZLibCompression(), MJCSSHConstants.SSH_TOKENIZER_STC);
    }  
    else if (PN_CO_STC.equals("none")) {
      debugOutput("Using none as compression for STC");
      tokenizer.setCompressionAlgorithm(new MJCSSHNoneCompression(), MJCSSHConstants.SSH_TOKENIZER_STC);
    }  
   
    debugOutput("STC Keys active");

    transportProtocolReady();
  }

  /** This method is invoked after the transport protocol is ready for data exchange.
      This method starts the authentication protocol.
  */
  private void transportProtocolReady() {
    if (!bTransportProtocolReady) {
      bTransportProtocolReady=true;
      sendUserAuthRequestMsg();
    }
  }

  /** Sends a service request message to the server with a specified service name.
      @param String serviceName: the service name to be requested
  */
  private void sendServiceRequestMsg(String payload) {
    String toSend=io.wByte(MJCSSHConstants.SSH_TRANSPORT_MSG_SERVICE_REQUEST);
    toSend=toSend.concat(io.wString(payload));
    send(toSend);
  }

  /** Processes the service accept message from the server.
      @param MJCSSHTransportPackage pack: the service accept message from the server
  */
  private void processServiceAcceptMsg(MJCSSHTransportPackage pack) {

    if (pack.payload.length()>1) {
      String serviceName=io.rString(pack.payload.substring(1));
      if (serviceName.equals("ssh-userauth")) {
        loginAttempt(LA_REQ_ACCEPT);
      }
      else {
        sendUnimplementedMsg(pack);
        closeConnection();
      }
    }
    else {
      loginAttempt(LA_REQ_ACCEPT);
    }
  }

  private void processDisconnectMsg(MJCSSHTransportPackage pack) {
    writeToReader("Disconnected. Reason="+io.rString((pack.payload).substring(5))+CRLF);
  }

  private void closeConnection() {
    debugOutput("Closing conn...");
    String temp=io.wByte(MJCSSHConstants.SSH_TRANSPORT_MSG_DISCONNECT);
    temp=temp.concat(io.wUint32(MJCSSHConstants.SSH_TRANSPORT_DISCONNECT_BY_APPLICATION));
    temp=temp.concat(io.wString("Requested by user."));
    temp=temp.concat(io.wString("en"));
    send(temp);
  }

  // *******************************************************************************************
  // Authentication protocol
  // *******************************************************************************************

  /** Sends the user authentication request to the server.
  */
  private void sendUserAuthRequestMsg() {
    loginAttempt(LA_REQ);
  }

  /** Processes the user authentication failure message from the server.
      @param MJCSSHTransportPackage pack: the user authentication failure fromt the server
  */
  private void processUserAuthFailureMsg(MJCSSHTransportPackage pack) {
    writeToReader(UO_INCORRECTLOGIN);
    String possibleContinuation=io.rString(pack.payload.substring(1));
    if (possibleContinuation.indexOf("password")!=-1) {
      loginAttempt(LA_REQ_ACCEPT);
    }
    else {
      writeToReader(UO_CANNOTCONTINUE);
      closeConnection();
    }
  }

  /** Login attempt manager method.
      @param int state: where the login attempt cycle is
  */
  private void loginAttempt(int state) {

    String toSend;

    switch (state) {
      case LA_REQ :
        sendServiceRequestMsg("ssh-userauth");
        break;
      case LA_REQ_ACCEPT :
        UIS_username=new String();
        UIS_password=new String();
        UII_username=true;
        UII_password=false;
        writeToReader(UO_USERNAME);
        break;
    }
  }

  /** Parses user input for username and password entering.
      @param String s: the String to be parsed.
      @return boolean:
  */
  private boolean parseUI_UP(String s) {
    int i;
    char c;
    char ca[]=new char[1];

    // username
    if (UII_username) {
      for (i=0; i<s.length(); i++) {
        c=s.charAt(i);
        ca[0]=c;

        if ((c>32) && (c<127)) { // chars
          if (UIS_username.length()<UI_USERNAME_LENGTH) {
            UIS_username=UIS_username.concat(new String(ca));
          }
          refreshUILine(UO_USERNAME, UIS_username, true);
        }
        else if ((c==8) || (c==127)) { // DEL
          if (UIS_username.length()>0) {
            UIS_username=UIS_username.substring(0,UIS_username.length()-1);
          }
          refreshUILine(UO_USERNAME, UIS_username, true);
        }
        else if (c==10) {
          if (UII_password==false) {
            writeToReader(CRLF+UO_PASSWORD);
            UII_username=false;
            UII_password=true;
          }
        }
        else if (c==13) {
          if (UII_password==false) {
            writeToReader(CRLF+UO_PASSWORD);
            UII_username=false;
            UII_password=true;
          }
        }
        else {
          break;
        }
      }
    }

    // password
    else if (UII_password) {
      for (i=0; i<s.length(); i++) {
        c=s.charAt(i);
        ca[0]=c;

        if ((c>32) && (c<127)) { // chars
          if (UIS_password.length()<UI_PASSWORD_LENGTH) {
            UIS_password=UIS_password.concat(new String(ca));
          }
          refreshUILine(UO_PASSWORD, UIS_password, false);
        }
        else if ((c==8) || (c==127)) { // DEL
          if (UIS_password.length()>0) {
            UIS_password=UIS_password.substring(0,UIS_password.length()-1);
          }
          refreshUILine(UO_PASSWORD, UIS_password, false);
        }
        else if (c==10) {
          if (UII_password==true) {
            UII_password=false;
            writeToReader(CRLF);
            sendLoginAttemptMsg();
          }
        }
        else if (c==13) {
          if (UII_password==true) {
            UII_password=false;
            writeToReader(CRLF);
            sendLoginAttemptMsg();
          }
        }
        else {
          break;
        }
      }
    }
    return true;
  }

  /** Refreshes the current line to contain the the data received as parameters.
      @param String banner: the fixed part of the line
      @param String ui: the user input
      @param boolean type: true-normal, false-password
  */
  private void refreshUILine(String banner, String ui, boolean type) {

    int i;
    String data=new String();

    data=data.concat(""+((char)13)); // HOME
    data=data.concat(""+((char)27)+((char)91)+((char)75)); // Erase EOL
    data=data.concat(banner);
    if (type) {
      data=data.concat(ui);
    }
    else {
      for (i=0; i<ui.length(); i++) data=data.concat("*");
    }
    writeToReader(data);
  }

  /** Sends a password type login attempt to the server.
  */
  private void sendLoginAttemptMsg() {
    String userAuth=io.wByte(MJCSSHConstants.SSH_USERAUTH_MSG_USERAUTH_REQUEST);
    userAuth=userAuth.concat(io.wString(UIS_username));
    userAuth=userAuth.concat(io.wString("ssh-connection"));
    userAuth=userAuth.concat(io.wString("password"));
    userAuth=userAuth.concat(io.wBoolean(false));
    userAuth=userAuth.concat(io.wString(UIS_password));
    send(userAuth);
  }

  private void processUserAuthSuccessMsg(MJCSSHTransportPackage pack) {
    authenticationProtocolReady();
  }

  /** This method is invoked after the authentication protocol is ready for data exchange.
      This method starts the connection protocol.
  */
  private void authenticationProtocolReady() {
    if (!bAuthenticationProtocolReady) {
      bAuthenticationProtocolReady=true;
      sendChannelOpenRequestMsg();
    }
  }

  // *******************************************************************************************
  // Connection protocol
  // *******************************************************************************************

  /** Sends the channel open request message to the server.
  */
  private void sendChannelOpenRequestMsg() {
    String openChannel=io.wByte(MJCSSHConstants.SSH_CONNECTION_MSG_CHANNEL_OPEN);
    openChannel=openChannel.concat(io.wString("session"));
    openChannel=openChannel.concat(io.wUint32(clientChannelNo));
    openChannel=openChannel.concat(io.wUint32(CHANNEL_WINDOW_SIZE));
    openChannel=openChannel.concat(io.wUint32(CHANNEL_MAX_PACKET_LENGTH));
    channelWindowSTC=CHANNEL_WINDOW_SIZE;
    maxPacketSizeSTC=CHANNEL_MAX_PACKET_LENGTH;
    send(openChannel);
  }

  /** Process the channel open confirmation message from the server.
      @param MJCSSHTransportPackage pack: the channel open confirmation message from the server
  */
  private void processChannelOpenConfirmationMsg(MJCSSHTransportPackage pack) {
    Vector data;
    data=io.parsePayload(pack.payload, MJCSSHConstants.CONN_CHANNEL_OPEN_CONFIRM_MSG_DATA_DESC);

    int recipent;
    recipent=((BigInteger)(data.elementAt(MJCSSHConstants.DI_CONN_CHANNEL_OPEN_CONFIRM_RECIPENT))).intValue();
    serverChannelNo=((BigInteger)(data.elementAt(MJCSSHConstants.DI_CONN_CHANNEL_OPEN_CONFIRM_SENDER))).intValue();
    channelWindowCTS=((BigInteger)(data.elementAt(MJCSSHConstants.DI_CONN_CHANNEL_OPEN_CONFIRM_INITIAL_WINDOW))).intValue();
    maxPacketSizeCTS=((BigInteger)(data.elementAt(MJCSSHConstants.DI_CONN_CHANNEL_OPEN_CONFIRM_MAXIMUM_PACKET))).intValue();

    sendReqPtyMsg();
  }

  /** Sends request for a pseudo terminal.
  */
  private void sendReqPtyMsg() {
    String pty=io.wByte(MJCSSHConstants.SSH_CONNECTION_MSG_CHANNEL_REQUEST);
    pty=pty.concat(io.wUint32(serverChannelNo));
    pty=pty.concat(io.wString("pty-req"));
    pty=pty.concat(io.wBoolean(true));
    pty=pty.concat(io.wString("vt220"));
    pty=pty.concat(io.wUint32(80));
    pty=pty.concat(io.wUint32(25));
    pty=pty.concat(io.wUint32(0));
    pty=pty.concat(io.wUint32(0));
    pty=pty.concat(io.wString(""));
    send(pty);
    sendShellReqMsg();
  }

  /** Sends request for shell.
  */
  private void sendShellReqMsg() {
    String shell=io.wByte(MJCSSHConstants.SSH_CONNECTION_MSG_CHANNEL_REQUEST);
    shell=shell.concat(io.wUint32(serverChannelNo));
    shell=shell.concat(io.wString("shell"));
    shell=shell.concat(io.wBoolean(false));
    send(shell);
  }

  /** Processes the channel success message.
      @param MJCSSHTransportPackage pack: the channel success message
  */
  private void processChannelSuccessMsg(MJCSSHTransportPackage pack) {
    connected=true;
  }

  /** Processes the channel data message.
      @param MJCSSHTransportPackage pack: the channel data message
  */
  private void processChannelDataMsg(MJCSSHTransportPackage pack) {
    String data=io.rString(pack.payload.substring(5));
    writeToReader(data);
    addNewWindowSize(data.length());
  }

  /** Increases the servers' window size by the specified amount.
      @param int amount: the amount with which the serves' window size will be increased
  */
  private void addNewWindowSize(int len) {
    String adjust=io.wByte(MJCSSHConstants.SSH_CONNECTION_MSG_CHANNEL_WINDOW_ADJUST);
    adjust=adjust.concat(io.wUint32(serverChannelNo));
    adjust=adjust.concat(io.wUint32(len));
    send(adjust);
  }

  /** Processes the window adjust message.
      @param MJCSSHTransportPackage pack: the window adjust message
  */
  private void processWindowAdjustMsg(MJCSSHTransportPackage pack) {
    channelWindowCTS+=io.rUint32Int(pack.payload.substring(5));
  }

  /** Processes the channel request message.
      @param MJCSSHTransportPackage pack: the channel request message
  */
  private void processChannelRequestMsg(MJCSSHTransportPackage pack) {
    writeToReader("Channel request."+io.rString(pack.payload.substring(5))+CRLF);
  }

  /** Processes the channel closed message.
      @param MJCSSHTransportPackage pack: the channel closed message
  */
  private void processChannelClosedMsg(MJCSSHTransportPackage pack) {
    writeToReader("Channel closed."+CRLF);
    closeConnection();
  }

  /** Sends the specified String as channel data.
      @param String s: the data to be sent through the channel
  */
  private void sendChannelDataMsg(String data) {
    String pack=io.wByte(MJCSSHConstants.SSH_CONNECTION_MSG_CHANNEL_DATA);
    pack=pack.concat(io.wUint32(serverChannelNo));
    pack=pack.concat(io.wString(data));
    send(pack);
  }

  // *******************************************************************************************
  // Stream processing
  // *******************************************************************************************

  /** Sends the contents of the buffer maintained as an MJCWriteable to the layer under.
  */
  private synchronized void sendWriteableBuffer() {
    sendChannelDataMsg(writeableBuf.emptyBuffer());
  }

  /** Adds data to be read out by the MJCReader from this layer as MJCReadable.
      @param String s: the data to be read out
  */
  private synchronized void writeToReader(String s) {
    readableBuf.addToBuffer(s);
    reader.readEvent(this);
  }

  /** Sends a String as payload for an SSH package via the under layer.
      @param String s: the payload to be send
  */
  private void send(String pack) {

    if ((connected) && (channelWindowCTS<pack.length())) {
      //--
      return;
    }

    synchronized (this) {
      pack=tokenizer.createPackage(pack);
      // System.err.println("Sending "+pack.length()+" bytes");

      writeable.write(pack);
      writeable.flush();
    }
  }

  private void dumpString(String d, PrintStream p) {
    p.println("LEN="+d.length());
    p.println(d);
    for (int i=0; i<d.length(); i++) {
      if (((int)d.charAt(i))>32) {
        p.println(""+i+" "+(int)d.charAt(i)+" = "+d.charAt(i));
      }
      else {
        p.println(""+i+" "+(int)d.charAt(i));
      }
    }
    p.println("----------------------------------------------------------");
  }

  private void debugOutput(String s) {
    System.err.println("SSH-TR>"+s);
    writeToReader("SSH-TR>"+s+CRLF);
  }


}
