/************************************************************************
*
* Copyright (C) 2001 by 
* Kerenyi Kristof, Toth Gergely
*
* Team Monica Javacom, Package version 0.2
*
* Report any comments to javacom@tg.sch.bme.hu
* 
* Available under the GPL.
* 
* For more details about the copyright terms and conditions read the
* GNU Public License (GPL) available under www.gnu.org/copyleft/gpl.txt.
* 
*********************************************************************** */


package com.monica.javacom.ssh;

import com.monica.javacom.interfaces.*;
import com.monica.javacom.ssh.*;

import java.util.*;
import java.math.*;

/** Class for providing functionality to create and receive SSH2 packets. This class
    is able to create according to the specified MAC and Ciphering algorithms packets
    from the given payload, and is also able to read SSH2 packets from the given
    String stream, by using MAC and Ciphering.
*/    
public class MJCSSHTransportPackageTokenizer {

  private MJCSSHMACAlgorithm macAlgorithmCTS, macAlgorithmSTC;
  private MJCSSHCipherAlgorithm cipherAlgorithmCTS, cipherAlgorithmSTC;
  private MJCSSHCompressionAlgorithm compressionAlgorithmCTS, compressionAlgorithmSTC;

  private String lastError;
  private String buffer;

  private MJCSSHIO io;

  private Vector packages;

  private int minProcessSizeCTS, minProcessSizeSTC;
  private int macLengthCTS, macLengthSTC;

  private String IV_CTS, IV_STC, EK_CTS, EK_STC, IK_CTS, IK_STC;

  private int sequenceNumberCTS, sequenceNumberSTC;

  private static final String ERR_NO_PACKAGE="There is no package to retrieve from the SSH transport layer.";

  /** Default constructor. */
  public MJCSSHTransportPackageTokenizer() {
    io=new MJCSSHIO();
    packages=new Vector();
    lastError=new String();
    buffer=new String();
    IV_CTS=new String();
    IV_STC=new String();
    EK_CTS=new String();
    EK_STC=new String();
    IK_CTS=new String();
    IK_STC=new String();
    sequenceNumberCTS=0;
    sequenceNumberSTC=0;
    minProcessSizeCTS=8;
    minProcessSizeSTC=8;
    macLengthCTS=0;
    macLengthSTC=0;
  }

  // true=CTS, false=STC
  /** Sets the keys for encryption (iv, key) and integrity. This method
      has to be called before the setCipherAlgorithm, setCompressionAlgorithm
      and setMACAlgorithm methods.
      @param String iv: the initialization vector for ciphering
      @param String ek: the encryption key
      @param String ik: the integrity key (for MAC)
      @param boolean dir: true for client to server, false for server to client
  */      
  public void setKeys(String iv, String ek, String ik, boolean dir) {
    if (dir) {
      IV_CTS=iv;
      EK_CTS=ek;
      IK_CTS=ik;
    }
    else {
      IV_STC=iv;
      EK_STC=ek;
      IK_STC=ik;
    }
  }

  // true=CTS, false=STC
  /** Sets the MAC algorithm that has to be used.
      @param MJCSSHMACAlgorithm: the MAC algorithm to be used
      @param boolean dir: true for client to server, false for server to client
  */      
  public void setMACAlgorithm(MJCSSHMACAlgorithm ma, boolean dir) {
    if (dir) {
      macAlgorithmCTS=ma;
      macLengthCTS=macAlgorithmCTS.getMACLength();
    }
    else {
      macAlgorithmSTC=ma;
      macLengthSTC=macAlgorithmSTC.getMACLength();
    }
  }

  // true=CTS, false=STC
  /** Sets the Cipheralgorithm that has to be used.
      @param MJCSSCipherAlgorithm: the Cipher algorithm to be used
      @param boolean dir: true for client to server, false for server to client
  */      
  public void setCipherAlgorithm(MJCSSHCipherAlgorithm ca, boolean dir) {
    if (dir) {
      cipherAlgorithmCTS=ca;
      cipherAlgorithmCTS.setKeys(EK_CTS, IV_CTS);
      if (cipherAlgorithmCTS.getCipherBlockSize()>minProcessSizeCTS) {
        minProcessSizeCTS=cipherAlgorithmCTS.getCipherBlockSize();
      }
    }
    else {
      cipherAlgorithmSTC=ca;
      cipherAlgorithmSTC.setKeys(EK_STC, IV_STC);
      if (cipherAlgorithmSTC.getCipherBlockSize()>minProcessSizeSTC) {
        minProcessSizeSTC=cipherAlgorithmSTC.getCipherBlockSize();
      }
    }
  }

  // true=CTS, false=STC
  /** Sets the Compression algorithm that has to be used.
      @param MJCSSHCompressionAlgorithm: the MAC algorithm to be used
      @param boolean dir: true for client to server, false for server to client
  */      
  public void setCompressionAlgorithm(MJCSSHCompressionAlgorithm ca, boolean dir) {
    if (dir) {
      compressionAlgorithmCTS=ca;
    }
    else {
      compressionAlgorithmSTC=ca;
    }
  }

  /** This method tells this object that at the to be processsed stream
      new data has arrived. 
      @param String s: the arrived data
      @throws Exception: if there has been an Exception during the processing of
        the arrived data
  */      
  public synchronized void arrived(String s) throws Exception {
    buffer=buffer.concat(s);
    processBuffer();
  }

  /** Returns true if there are processed packets that can be retrieved.
      @return boolean: true if there processed packets that can be retrieved
  */      
  public synchronized boolean isAvailable() {
    return (packages.size()>0);
  }

  /** Returns the next available packet from the processed packets. 
      @return MJCSSHTransportPackage: the next packet from the processed packets
        or null if there aren't any
  */	
  public synchronized MJCSSHTransportPackage getNextPackage() {
    if (packages.size()>0) {
      MJCSSHTransportPackage p=((MJCSSHTransportPackage)packages.elementAt(0));
      packages.removeElementAt(0);
      return p;
    }
    else {
      lastError=ERR_NO_PACKAGE;
      return null;
    }
  }

  /** Returns the description of the last occured error. By calling this method
      the error description will be lost.
      @return String: the last occured error
  */      
  public String getLastError() {
    String temp=lastError;
    lastError=new String();
    return temp;
  }

  private void processBuffer() throws Exception {
    String pckl_e, pckl_d, p_e, p_d;
    BigInteger pckl, padl;
    int pckl_int, padl_int, payload_l;
    MJCSSHTransportPackage pack;
    String unencryptedPacket;
    Object o;

    while (buffer.length()>minProcessSizeSTC) {
      pckl_e=buffer.substring(0,minProcessSizeSTC);
      o=cipherAlgorithmSTC.getState();

      pckl_d=cipherAlgorithmSTC.decode(pckl_e);
      pckl=io.rUint32(pckl_d);
      pckl_int=pckl.intValue();

      if (pckl_int<0) {
        throw new Exception("smaller: ="+io.print(pckl_d.substring(0,4))+", intvalue="+pckl_int);
      }

      if (buffer.length()>=(4+pckl_int+macLengthSTC)) { // whole package is here
        p_e=buffer.substring(minProcessSizeSTC,4+pckl_int);
        p_d=pckl_d.concat(cipherAlgorithmSTC.decode(p_e)); //decrypted packet data

        pack=new MJCSSHTransportPackage();

        pack.packet_length=pckl;
        pack.padding_length=io.rByte(p_d.substring(4,5));
        padl_int=pack.padding_length.intValue();
        payload_l=pckl_int-padl_int-1;

        pack.payload=p_d.substring(5,5+payload_l);
        
        // decompressing payload
        pack.payload=compressionAlgorithmSTC.decompress(pack.payload);

        pack.random_padding=p_d.substring(5+payload_l,5+payload_l+padl_int);

        pack.mac=buffer.substring(pckl_int+4,pckl_int+4+macLengthSTC);

        unencryptedPacket=io.wUint32(sequenceNumberSTC);
        pack.sequenceNumber=sequenceNumberSTC;

        sequenceNumberSTC++;
        unencryptedPacket=unencryptedPacket.concat(p_d);

//        System.err.println(""+(int)(pack.payload.charAt(0))+", "+(sequenceNumberSTC-1));
        if (!macAlgorithmSTC.authenticate(unencryptedPacket,pack.mac,IK_STC)) {
          System.err.println("REC PACKauth Failed!");
          throw new Exception("REC PACKauth Failed!");
        }

        buffer=buffer.substring(4+pckl_int+macLengthSTC);

        packages.addElement(pack);
      }
      else {
        cipherAlgorithmSTC.setState(o);
        break;
      }
    }

  }

  /** Creates an encrypted, MAC-ed, compressed SSH2 transport packet from the supplied payload
      and returns it as a String.
      @param String payload: the packet payload
      @return String: the full SSH2 transport packet
  */      
  public String createPackage(String payload) {

    // compressing payload;    
    payload=compressionAlgorithmCTS.compress(payload);
    
    int len_without_mac;
    int padding_len;
    String result;
    String mac;
    char padding[];

    len_without_mac=16;
    padding_len=len_without_mac-5-payload.length();
    if (padding_len<4) padding_len=4;
    len_without_mac=padding_len+payload.length()+5;
    while ((len_without_mac%minProcessSizeCTS)!=0) len_without_mac++;
    padding_len=len_without_mac-5-payload.length();

    result=io.wUint32(len_without_mac-4);
    result=result.concat(io.wByte(padding_len));
    result=result.concat(payload);

    padding=new char[padding_len];
    for (int i=0; i<padding_len; i++) {
      padding[i]=(char)(Math.random()*255.0);
    }

    result=result.concat(new String(padding));

    mac=macAlgorithmCTS.createMAC((io.wUint32(sequenceNumberCTS)).concat(result),IK_CTS);
    sequenceNumberCTS++;

    result=cipherAlgorithmCTS.encode(result);

    result=result.concat(mac);

    return result;
  }

}